"""Interoperates @dataclass with json objects"""

from jsondaora.dataclasses import asdataclass
from jsondaora.decorator import jsondaora
from jsondaora.schema import (
    IntegerField,
    StringField,
    integer,
    jsonschema_asdataclass,
    string,
)
from jsondaora.serializers import dataclass_asjson, typed_dict_asjson
from jsondaora.typed_dict import as_typed_dict, as_typed_dict_field


__version__ = '0.12.0'


__all__ = [
    'jsondaora',
    'StringField',
    'IntegerField',
    'asdataclass',
    'as_typed_dict',
    'as_typed_dict_field',
    'dataclass_asjson',
    'typed_dict_asjson',
    'integer',
    'string',
    'jsonschema_asdataclass',
]



import dataclasses
from datetime import datetime
from logging import getLogger
from typing import (  # type: ignore
    TYPE_CHECKING,
    Any,
    Dict,
    Optional,
    Set,
    Type,
    Union,
    _GenericAlias,
)

import orjson

from .exceptions import DeserializationError
from .fields import DeserializeFields


logger = getLogger(__name__)

_ERROR_MSG = 'Invalid type={annotation} for field={field_name}'


def deserialize_jsondict_fields(
    jsondict: Dict[str, Any],
    cls: Type[Any],
    skip_fields: Set[str] = set(),
    encode_field_name: bool = False,
) -> Dict[str, Any]:
    custom_fields = DeserializeFields.get_fields(cls)
    all_fields = dataclasses.fields(cls)
    deserialized = {}
    fields = custom_fields if custom_fields else all_fields
    skipped_fields = set()

    if skip_fields:
        new_fields = set()
        for f in fields:
            if f.name in skip_fields:
                skipped_fields.add(f)
            else:
                new_fields.add(f)
        fields = new_fields

    if encode_field_name:
        for field in fields:
            value = jsondict.get(field.name.encode())
            deserialized[field.name] = deserialize_field(
                field_name=field.name,
                field_type=field.type,
                field_default=field.default,
                value=value,
                cls=cls,
            )
    else:
        for field in fields:
            value = jsondict.get(field.name)
            deserialized[field.name] = deserialize_field(
                field_name=field.name,
                field_type=field.type,
                field_default=field.default,
                value=value,
                cls=cls,
            )

    if custom_fields or skipped_fields:
        for field in (
            set(all_fields) - custom_fields if custom_fields else set()
        ).union(skipped_fields):
            deserialized[field.name] = jsondict[field.name]

    return deserialized


if TYPE_CHECKING:
    _Field = dataclasses.Field[Any]
else:
    _Field = dataclasses.Field


def deserialize_field(
    field_name: str,
    field_type: Any,
    value: Any,
    cls: Optional[Type[Any]] = None,
    field_default: Any = dataclasses.MISSING,
) -> Any:
    try:
        if hasattr(field_type, '__get_dynamic_type__'):
            dynamic_type = field_type.__get_dynamic_type__(value)
            return deserialize_field(
                field_name=field_name,
                field_type=dynamic_type,
                value=value,
                cls=cls,
                field_default=field_default,
            )

        elif isinstance(value, dict) and dataclasses.is_dataclass(field_type):
            value = deserialize_jsondict_fields(value, field_type)
            return field_type(**value)

        elif (isinstance(value, bytes) or isinstance(value, str)) and (
            dataclasses.is_dataclass(field_type)
            or (
                isinstance(field_type, _GenericAlias)
                and (
                    field_type.__origin__ is dict
                    or field_type.__origin__ is list
                    or field_type.__origin__ is tuple
                )
            )
        ):
            return deserialize_field(
                field_name=field_name,
                field_type=field_type,
                value=orjson.loads(value),
                cls=cls,
                field_default=field_default,
            )

        elif isinstance(field_type, _GenericAlias):
            return _deserialize_generic_type(field_type, field_name, value)

        if field_type is Any:
            return value

        elif isinstance(value, field_type):
            return value

        elif (
            isinstance(field_type, type)
            and issubclass(field_type, str)
            and isinstance(value, bytes)
        ):
            return value.decode()

        elif (
            isinstance(field_type, type)
            and issubclass(field_type, bytes)
            and isinstance(value, str)
        ):
            return value.encode()

        elif field_type is datetime:
            return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')

        if field_type is bool:
            return value in (b'1', 1, '1', 't', 'true', 'y', 'yes')

        elif value is None and field_default is None:
            return None

        elif field_default is not dataclasses.MISSING and value is None:
            return field_type(field_default)

        if value is not None:
            return field_type(value)

        raise DeserializationError(
            field_name, field_type, field_default, value, cls
        )

    except (TypeError, ValueError) as error:
        raise DeserializationError(
            field_name, field_type, field_default, value, cls
        ) from error


def _deserialize_generic_type(
    generic: Any, field_name: str, value: Any
) -> Any:
    try:
        return _DESERIALIZERS_MAP[generic.__origin__](
            generic, field_name, value
        )
    except KeyError:
        raise DeserializationError(
            _ERROR_MSG.format(annotation=generic, field_name=field_name)
        )


def _deserialize_union(generic: Any, field_name: str, value: Any) -> Any:
    nullable = False

    for type_ in generic.__args__:
        if type_ is not None and type_ is not type(None):  # noqa
            try:
                return deserialize_field(
                    field_name=field_name,
                    field_type=type_,
                    value=value,
                    cls=generic,
                )
            except DeserializationError:
                continue
        else:
            nullable = True

    if nullable:
        return None

    raise DeserializationError(
        _ERROR_MSG.format(annotation=generic, field_name=field_name)
    )


def _deserialize_list(annotation: Any, field_name: str, values: Any) -> Any:
    try:
        return [
            deserialize_field(
                field_name=field_name,
                field_type=annotation.__args__[0],
                value=value,
                cls=annotation,
            )
            for value in values
        ]
    except TypeError as err:
        raise DeserializationError(
            _ERROR_MSG.format(annotation=annotation, field_name=field_name)
        ) from err


def _deserialize_tuple(annotation: Any, field_name: str, values: Any) -> Any:
    try:
        if annotation.__args__[-1] is ...:
            return tuple(
                deserialize_field(
                    field_name=field_name,
                    field_type=annotation.__args__[0],
                    value=value,
                    cls=annotation,
                )
                for value in values
            )

        elif len(annotation.__args__) == len(values):
            return tuple(
                deserialize_field(
                    field_name=field_name,
                    field_type=annotation.__args__[i],
                    value=value,
                    cls=annotation,
                )
                for i, value in enumerate(values)
            )

        else:
            raise DeserializationError(
                _ERROR_MSG.format(annotation=annotation, field_name=field_name)
            )

    except TypeError as err:
        raise DeserializationError(
            _ERROR_MSG.format(annotation=annotation, field_name=field_name)
        ) from err


def _deserialize_set(annotation: Any, field_name: str, values: Any) -> Any:
    try:
        return set(
            deserialize_field(
                field_name=field_name,
                field_type=annotation.__args__[0],
                value=value,
                cls=annotation,
            )
            for value in values
        )
    except TypeError as err:
        raise DeserializationError(
            _ERROR_MSG.format(annotation=annotation, field_name=field_name)
        ) from err


def _deserialize_dict(annotation: Any, field_name: str, values: Any) -> Any:
    new_dict: Dict[Any, Any] = {}

    try:
        for i, (key, value) in enumerate(values.items()):
            new_dict[
                deserialize_field(
                    field_name=f'{field_name}.key{i}',
                    field_type=annotation.__args__[0],
                    value=key,
                    cls=annotation,
                )
            ] = deserialize_field(
                field_name=f'{field_name}.value{i}',
                field_type=annotation.__args__[1],
                value=value,
                cls=annotation,
            )

    except (TypeError, AttributeError) as err:
        raise DeserializationError(
            _ERROR_MSG.format(annotation=annotation, field_name=field_name)
        ) from err

    else:
        return new_dict


_DESERIALIZERS_MAP = {
    Union: _deserialize_union,
    list: _deserialize_list,
    tuple: _deserialize_tuple,
    set: _deserialize_set,
    dict: _deserialize_dict,
}



import dataclasses
from types import MethodType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from dictdaora import DictDaora

from .exceptions import DeserializationError


class StringField(DictDaora):
    max_length: int
    min_length: int
    value: str
    validate: Callable[[Type['StringField'], str], None]

    def __init__(self, value: Any):
        self.value = value.decode() if isinstance(value, bytes) else str(value)
        type(self).validate(self.value)  # type: ignore

    def __init_subclass__(
        cls, min_length: Optional[int] = None, max_length: Optional[int] = None
    ) -> None:
        if max_length is None and min_length is not None:
            cls.validate = MethodType(validate_min_length, cls)

        elif max_length is not None and min_length is None:
            cls.validate = MethodType(validate_max_length, cls)

        elif max_length is not None and min_length is not None:
            cls.validate = MethodType(validate_min_length_max_length, cls)

        elif min_length is None and max_length is None:
            cls.validate = MethodType(lambda c, v: v, cls)

        if min_length is not None:
            cls.min_length = min_length

        if max_length is not None:
            cls.max_length = max_length


class IntegerField(DictDaora):
    maximum: int
    minimum: int
    value: int
    validate: Callable[[Type['IntegerField'], int], None]

    def __init__(self, value: Any):
        self.value = int(value)
        type(self).validate(self.value)  # type: ignore

    def __init_subclass__(
        cls, minimum: Optional[int] = None, maximum: Optional[int] = None
    ) -> None:
        if maximum is None and minimum is not None:
            cls.validate = MethodType(validate_minimum, cls)

        elif maximum is not None and minimum is None:
            cls.validate = MethodType(validate_maximum, cls)

        elif maximum is not None and minimum is not None:
            cls.validate = MethodType(validate_minimum_maximum, cls)

        elif maximum is None and minimum is None:
            cls.validate = MethodType(lambda c, v: v, cls)

        if minimum is not None:
            cls.minimum = minimum

        if maximum is not None:
            cls.maximum = maximum


def validate_minimum(cls: Type[IntegerField], value: int) -> None:
    if not cls.minimum <= value:
        raise DeserializationError(
            f'Invalid minimum integer value: {cls.minimum} < {value}'
        )


def validate_maximum(cls: Type[IntegerField], value: int) -> None:
    if not cls.maximum >= value:
        raise DeserializationError(
            f'Invalid maximum integer value: {value} < {cls.maximum}'
        )


def validate_minimum_maximum(cls: Type[IntegerField], value: int) -> None:
    if not cls.minimum <= value <= cls.maximum:
        raise DeserializationError(
            f'Invalid minimum and maximum integer value: '
            f'{cls.minimum} < {value} < {cls.maximum}'
        )


def validate_min_length(cls: Type[StringField], value: str) -> None:
    if not cls.min_length <= len(value):
        raise DeserializationError(
            f'Invalid min_length string value: {cls.min_length} < {len(value)}'
        )


def validate_max_length(cls: Type[StringField], value: str) -> None:
    if not cls.max_length >= len(value):
        raise DeserializationError(
            f'Invalid max_length string value: {len(value)} < {cls.max_length}'
        )


def validate_min_length_max_length(cls: Type[StringField], value: str) -> None:
    if not cls.min_length <= len(value) <= cls.max_length:
        raise DeserializationError(
            f'Invalid min_length and max_length string value: '
            f'{cls.min_length} < {len(value)} < {cls.max_length}'
        )


def string(
    min_length: Optional[int] = None, max_length: Optional[int] = None
) -> Type[StringField]:
    min_length_str = '' if min_length is None else str(min_length)
    max_length_str = '' if max_length is None else str(max_length)
    cls_name = f'String{min_length_str}{max_length_str}'
    return type(
        cls_name,
        (StringField,),
        {'min_length': min_length, 'max_length': max_length},
    )


def integer(
    minimum: Optional[int] = None, maximum: Optional[int] = None
) -> Type[IntegerField]:
    minimum_str = '' if minimum is None else str(minimum)
    maximum_str = '' if maximum is None else str(maximum)
    cls_name = f'String{minimum_str}{maximum_str}'
    return type(
        cls_name, (IntegerField,), {'minimum': minimum, 'maximum': maximum},
    )


def jsonschema_asdataclass(
    id_: str, schema: Dict[str, Any], bases: Tuple[type, ...] = ()
) -> Type[Any]:
    return dataclasses.make_dataclass(
        id_,
        [
            (
                prop_name,
                Optional[
                    jsonschema_asdataclass(f'{id_}_{prop_name}', prop)  # noqa
                ]
                if prop['type'] == 'object'
                else (
                    Optional[jsonschema_array(id_, prop_name, prop)]
                    if prop['type'] == 'array'
                    else Optional[SCALARS[prop['type']]]
                ),
                dataclasses.field(default=prop.get('default')),
            )
            for prop_name, prop in schema['properties'].items()
        ],
        bases=bases,
    )


def jsonschema_array(id_: str, prop_name: str, prop: Any) -> Any:
    DynamicType = (
        jsonschema_asdataclass(f'{id_}_{prop_name}', prop['items'])  # noqa
        if (array_type := prop['items']['type']) == 'object'  # noqa
        else jsonschema_array(id_, prop_name, prop['items'])  # noqa
        if array_type == 'array'  # noqa
        else SCALARS[array_type]
    )
    return List[DynamicType]  # type: ignore


SCALARS = {
    'boolean': bool,
    'string': str,
    'integer': int,
    'number': float,
}



import dataclasses
from enum import Enum
from typing import Any, Dict, Set, Type

from .deserializers import deserialize_jsondict_fields
from .exceptions import DeserializationError
from .fields import SerializeFields
from .serializers import dataclass_asjson


def asdataclass(
    jsondict: Dict[str, Any],
    cls: Type[Any],
    skip_fields: Set[str] = set(),
    encode_field_name: bool = False,
) -> Any:
    kwargs = deserialize_jsondict_fields(
        jsondict, cls, skip_fields, encode_field_name
    )

    try:
        return cls(**kwargs)
    except TypeError as error:
        raise DeserializationError(
            cls.__name__, jsondict, error, cls
        ) from error


def asdict(instance: Any, dumps_value: bool = False) -> Any:
    if isinstance(instance, list) or isinstance(instance, tuple):
        return [asdict(value) for value in instance]

    elif isinstance(instance, Enum):
        return instance.value

    elif isinstance(instance, dict):
        return {
            f: dataclass_asjson(value)
            if dumps_value
            and (
                dataclasses.is_dataclass(value)
                or isinstance(value, list)
                or isinstance(value, tuple)
                or isinstance(value, dict)
            )
            else asdict(value)
            for f, value in instance.items()
        }

    elif not dataclasses.is_dataclass(instance):
        return instance

    fields = SerializeFields.get_fields(type(instance))

    if not fields:
        fields = dataclasses.fields(instance)

    return {
        f.name: dataclass_asjson(value)
        if dumps_value
        and (
            dataclasses.is_dataclass(
                value := getattr(instance, f.name, f.default)  # noqa
            )
            or isinstance(value, list)
            or isinstance(value, tuple)
            or isinstance(value, dict)
        )
        else asdict(getattr(instance, f.name, f.default))
        for f in fields
    }
